// Clase OrdVect: aplicacin de mtodos de ordenamiento.
// Autor. Ing. Tymoschuk, Jorge 
/* OrdVect contiene el comportamiento necesario para ordenar
	un array lineal de Item's con diversos mtodos.
	La mayora de ellos son los llamados "bsicos".
	Los "mejorados" son: Peinado, Shell, Monticulo y Rapido.
	Item, en vez de ser una estructura, se define como una clase.
	Y entonces Item *a es una declaracin de un puntero a un array
	de objetos clase Item.
	El constructor de la clase OrdVect genera este array con tantos
	elementos como se le informa(Parmetro), llenandolo de nmeros
	generados aleatoriamente por la funcin random. */

typedef int Indice;
typedef int Clave;
#include <stdlib.h>
#include <iostream.h>

class Item {        //
public:
	Clave clave;
	int valor;
	char dato;
};

class OrdVect {
  Item   *a;
  Indice N;
  int    cent;
  void  Criba(Indice i, Indice de);
  void  Rapid(Indice iz, Indice de);
public:
		  OrdVect(Indice talle,int centinelas = 0);
		  ~OrdVect() { delete a; }
  Item& operator [](Indice i) { return a[i+cent]; }
  void  InsDirecta();
  void  InsDirecta2();
  void  InsBinaria();
  void  SelDirecta();
  void  Burbuja();
  void  Sacudida();
  void  Peinado();
  void  Shell();
  void  Monticulo();
  void  Rapido();
  Item  Encontrar(Indice k);
  void  MezclaDir();
  void  Imprim();
};

OrdVect::OrdVect(Indice talle,int centinelas) {
	 cent = centinelas; N = talle;
	 a = new Item [N+cent];
	 for(int i=0;i<N;i++)
		  a[i].clave=random(N);
}

void OrdVect::InsDirecta()       //definido a[N+1] por el centinela
{ Indice i,j; Item x;
  for (i = 2; i<=N; i++) {
    x = a[i]; a[0] = x;
    for (j = i-1; x.clave<a[j].clave; j--)
		a[j+1] = a[j];
    a[j+1] = x;  }
}

void OrdVect::InsDirecta2()      //definido a[N+1] por el centinela
{ Indice i = N-1;
  Item *cent = a, *x = &a[1], *p;
  while (i--) {
    p = x++; *cent = *x;
    while (cent->clave<p->clave)
      *(p+1) = *p--;
      *(p+1) = *cent;  }
}

void OrdVect::InsBinaria()       //definido a[N]
{ Indice i,j,iz,de,m; Item x;
  for (i = 1; i<N; i++) {
     x = a[i]; iz = 0; de = i-1;
     while (iz<=de) {
       m = (iz+de)/2;
       if (x.clave<a[m].clave) de = m-1; else iz = m+1;
     }
     for (j = i-1; j>=iz; j--) a[j+1] = a[j];
       a[iz] = x;
  }
}

void OrdVect::SelDirecta()
{ Indice i,j,k; Item min;
  for (i = 0; i<N-1; i++) {
    k = i; min = a[i];
    for (j = i+1; j<N; j++)
      if (a[j].clave<min.clave) {
	k = j; min = a[j];
      }
    a[k] = a[i]; a[i] = min;
  }
}

void OrdVect::Burbuja()
{ Indice i,j; Item aux;
  for (i = 1; i<N; i++)
    for (j = N-1; j>=i; j--)
      if (a[j-1].clave>a[j].clave) {
        aux = a[j-1]; a[j-1] = a[j]; a[j] = aux;
      }
}

void OrdVect::Sacudida()
{ Indice j,k = N-1,iz = 1,de = N-1; Item aux;
  do {
    for (j = de; j>=iz; j--)
      if (a[j-1].clave>a[j].clave) {
	aux = a[j-1]; a[j-1] = a[j]; a[j] = aux;
	k = j;
      }
    iz = k+1;
    for (j = iz; j<=de; j++)
    if (a[j-1].clave>a[j].clave) {
      aux = a[j-1]; a[j-1] = a[j]; a[j] = aux;
      k = j;
    }
    de = k-1;
  } while (iz<=de);
}

void OrdVect::Peinado()	//programa de ordenamiento por peinado
{ Indice i,j,paso = N; Item aux;
  int cambio;
  do {
    paso = (int)((float) paso/1.3);
    paso = paso>1 ? paso : 1;
    cambio = 0;
    for (i = 0; i<N-paso; i++)
      if (a[i].clave>a[j = i+paso].clave) {
	aux = a[i]; a[i] = a[j]; a[j] = aux; cambio = 1;
      }
  } while (cambio || paso>1);
}

#define	INTERVALO_0	9				 
#define	T               4

//definido a[N+INTERVALO_0]; debido a que los h[0] primeros componentes
//son centinelas (ver InsDirecta() e InsDirecta2())

void OrdVect::Shell()
{ Indice i,j,k,s; Item x;
  int m,h[T] = { INTERVALO_0, 5, 3, 1};
  for (m = 0; m<T; m++) {
    k = h[m]; s = h[0]-k-1;
    for (i = k+h[0]; i<N+INTERVALO_0; i++) {
      x = a[i]; j = i-k;
      s++;
      if (!(s-h[0])) s -= k;
      a[s] = x;
      while (x.clave<a[j].clave) {
		  a[j+k] = a[j]; j -= k;
      }
      a[j+k] = x;
    }
  }
}                               //note la similitud con insercin directa

void OrdVect::Monticulo()
{ Indice iz = N/2,de = N-1;
  Item x;
  while (iz--) Criba(iz,de);
  while (de) {
    x = a[0]; a[0] = a[de]; a[de] = x;
    de--; Criba(0,de);
  }
}

void OrdVect::Criba(Indice i, Indice de)
{ Indice j = 2*i+1;
  Item x = a[i];
  while (j<=de) {
    if (j<de && a[j].clave<a[j+1].clave) j++;
    if (x.clave>=a[j].clave) break;
    a[i] = a[j]; i = j; j = 2*j+1;
  }
  a[i] = x;
}

void OrdVect::Rapido()
{ Rapid(0,N-1); }

void OrdVect::Rapid(Indice iz, Indice de)
{ Indice i = iz,j = de;
  Item aux,x = a[(iz+de)/2];
  do {
    while (a[i].clave<x.clave) i++;
    while (x.clave<a[j].clave) j--;
    if (i<=j) {
      aux = a[i]; a[i] = a[j]; a[j] = aux;
      i++; j--;
    }
  } while (i<=j);
  if (iz<j) Rapid(iz,j);
  if (i<de) Rapid(i,de);
}

Item OrdVect::Encontrar(Indice k)
{ Indice iz = 0,de = N-1,i,j;	//Indice debe tener signo
  Item aux,x;
  while (iz<de) {
    x = a[k]; i = iz; j = de;
    do {
      while (a[i].clave<x.clave) i++;
      while (x.clave<a[j].clave) j--;
      if (i<=j) {
	aux = a[i]; a[i] = a[j]; a[j] = aux;
	i++; j--;
      }
    } while (i<=j);
    if (j<k) iz = i;
    if (k<i) de = j;
  }
  return a[k];
}

void OrdVect::MezclaDir()
{ Indice i,j,k,l,t;
  int h,m,p = 1,q,r,arriba = 1;
  do {	h = 1; m = N;
	 if (arriba) {     i = 0; j = N-1; k = N; l = 2*N-1;    }
	 else {   k = 0; l = N-1; i = N; j = 2*N-1;    }
	 do {
		q = m>=p ? p:m; m -= q;
		r = m>=p ? p:m; m -= r;
		while (q && r)
	if (a[i].clave<a[j].clave) {  a[k] = a[i]; k += h; i++; q--;}
		else {	 a[k] = a[j]; k += h; j--; r--;	}
		while (r)
{	a[k] = a[j]; k += h; j--; r--;      }
		while (q)
{	a[k] = a[i]; k += h; i++; q--;      }
		h = -h; t = k; k = l; l = t;
	 } while (m);
	 arriba = !arriba; p = 2*p;
  } while (p<N-1);
  if (!arriba) for (i = 0; i<N; i++) a[i] = a[i+N];
}

void OrdVect::Imprim()
{ int i;
	for (i=0;i<N;i++){
		if(i%10==0)cout << "\n";
		cout << a[i].clave  <<", ";
   }
}

void main(){
	cout << " Probando mtodos de ordenamiento\n";
	cout << "definidos en la clase OrdVect   \n";
	cout << "=================================\n";
	OrdVect v1(20),v2(30);

	cout << "\n\nImprimiendo v1 antes de su ordenamiento";
	v1.Imprim();

	cout << "\n\nImprimiendo v2 antes de su ordenamiento";
	v2.Imprim();

	v1.Burbuja();   //  Ordeno v1
	cout << "\n\nImprimiendo v1 ordenado por mtodo Burbuja";
	v1.Imprim();

	v2.Sacudida();  //  Ordeno v2
	cout << "\n\nImprimiendo v2 ordenado por mtodo Sacudida";
	v2.Imprim();
	cout << "\n\nTarea terminada ...";
}
